#ifndef DVM_DEV_H_INCLUDED
#define DVM_DEV_H_INCLUDED
#include "DVM.h"

typedef struct DVM_Array_tag DVM_Array;
typedef struct DVM_Context_tag DVM_Context;

typedef enum {
    DVM_SUCCESS = 1,
    DVM_ERROR
} DVM_ErrorStatus;

typedef DVM_Value DVM_NativeFunctionProc(DVM_VirtualMachine *dvm,
                                         DVM_Context *context,
                                         int arg_count, DVM_Value *args);

typedef void DVM_NativePointerFinalizeProc(DVM_VirtualMachine *dvm,
                                           DVM_Object *obj);

typedef struct {
    char                                *name;
    DVM_NativePointerFinalizeProc       *finalizer;
} DVM_NativePointerInfo;

typedef enum {
    DVM_INT_MESSAGE_ARGUMENT = 1,
    DVM_DOUBLE_MESSAGE_ARGUMENT,
    DVM_STRING_MESSAGE_ARGUMENT,
    DVM_CHARACTER_MESSAGE_ARGUMENT,
    DVM_POINTER_MESSAGE_ARGUMENT,
    DVM_MESSAGE_ARGUMENT_END
} DVM_MessageArgumentType;

typedef struct {
    char *format;
} DVM_ErrorDefinition;

typedef struct {
    DVM_ErrorDefinition *message_format;
} DVM_NativeLibInfo;

/* execute.c */
void DVM_add_native_function(DVM_VirtualMachine *dvm,
                             char *package_name, char *func_name,
                             DVM_NativeFunctionProc *proc, int arg_count,
                             DVM_Boolean is_method,
                             DVM_Boolean return_pointer);
/* nativeif.c */
DVM_ErrorStatus
DVM_array_get_int(DVM_VirtualMachine *dvm, DVM_ObjectRef array, int index,
                  int *value, DVM_ObjectRef *exception_p);
DVM_ErrorStatus
DVM_array_get_double(DVM_VirtualMachine *dvm, DVM_ObjectRef array,
                     int index, double *value,
                     DVM_ObjectRef *exception_p);
DVM_ErrorStatus
DVM_array_get_object(DVM_VirtualMachine *dvm, DVM_ObjectRef array, int index,
                     DVM_ObjectRef *value, DVM_ObjectRef *exception_p);
DVM_ErrorStatus
DVM_array_set_int(DVM_VirtualMachine *dvm, DVM_ObjectRef array, int index,
                  int value, DVM_ObjectRef *exception_p);
DVM_ErrorStatus
DVM_array_set_double(DVM_VirtualMachine *dvm, DVM_ObjectRef array, int index,
                     double value, DVM_ObjectRef *exception_p);
DVM_ErrorStatus
DVM_array_set_object(DVM_VirtualMachine *dvm, DVM_ObjectRef array, int index,
                     DVM_ObjectRef value, DVM_ObjectRef *exception_p);
int DVM_array_size(DVM_VirtualMachine *dvm, DVM_Object *array);
void
DVM_array_resize(DVM_VirtualMachine *dvm, DVM_Object *array, int new_size);
void DVM_array_insert(DVM_VirtualMachine *dvm, DVM_Object *array, int pos,
                      DVM_Value value);
void DVM_array_remove(DVM_VirtualMachine *dvm, DVM_Object *array, int pos);
int DVM_string_length(DVM_VirtualMachine *dvm, DVM_Object *string);
DVM_Value DVM_string_substr(DVM_VirtualMachine *dvm, DVM_Object *str,
                            int pos, int len);
int DVM_get_field_index(DVM_VirtualMachine *dvm, DVM_ObjectRef obj,
                        char *field_name);
void DVM_set_exception(DVM_VirtualMachine *dvm, DVM_Context *context,
                       DVM_NativeLibInfo *lib_info,
                       char *package_name, char *class_name,
                       int error_id, ...);

/* heap.c */
void DVM_check_gc(DVM_VirtualMachine *dvm);
DVM_ObjectRef
DVM_create_array_int(DVM_VirtualMachine *dvm, DVM_Context *context,
                     int size);
DVM_ObjectRef
DVM_create_array_double(DVM_VirtualMachine *dvm, DVM_Context *context,
                        int size);
DVM_ObjectRef
DVM_create_array_object(DVM_VirtualMachine *dvm,  DVM_Context *context,
                        int size);
void *DVM_object_get_native_pointer(DVM_Object *obj);
DVM_Boolean DVM_check_native_pointer_type(DVM_Object *native_pointer,
                                          DVM_NativePointerInfo *info);
void DVM_object_set_native_pointer(DVM_Object *obj, void *p);
DVM_ObjectRef
DVM_create_dvm_string(DVM_VirtualMachine *dvm, DVM_Context *context,
                      DVM_Char *str);
DVM_ObjectRef
DVM_create_class_object(DVM_VirtualMachine *dvm, DVM_Context *context,
                        int class_index);
DVM_ObjectRef
DVM_create_native_pointer(DVM_VirtualMachine *dvm, DVM_Context *context,
                          void *pointer, DVM_NativePointerInfo *info);

/* wchar.c */
char *DVM_wcstombs(const wchar_t *src);
DVM_Char *DVM_mbstowcs(const char *src);

#endif /* DVM_DEV_H_INCLUDED */
